static char rcsid[] = "@(#)$Id: rc_parse.c,v 2.6 2016/02/27 12:14:16 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.6 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *           or  Kari Hurtta <elm@elmme-mailer.org>
 ******************************************************************************
 *
 *  Moved do_rc() from lib/read.rc and splitted to do_rc_parse()
 *  and do_rc_process()
 * 
 *  Based on Elm 2.4 src/read_rc.c. It have following copyright:
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

#include "elm_defs.h"
#include "rc_imp.h"
#include "s_elm.h"
#include "save_opts.h"

DEBUG_VAR(Debug,__FILE__,"config");

static const unsigned char *cs2us P_((const char *str));
static const unsigned char *cs2us(str) 
     const char *str;
{
    return (const unsigned char *)str;
}

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

#define PARSED_RC_magic		0xF405

struct parsed_rc {
    unsigned short magic;   /*    PARSED_RC_magic */

    struct buffer_data {
	char * buffer;
	int   buffer_len;
	int lineno;
    } * head_comments;
    int head_comments_len;

    struct parsed_line {
	struct rc_save_info_rec * record;
	
	char   * current_tag;

	char * rest;
	int    restlen;

	int lineno;

	struct hist_record *hist_record;

	unsigned int is_cline: 1;
	unsigned int negate: 1;
    } * setting_lines;
    int setting_lines_len;
};

void free_parsed_rc(rc)
     struct parsed_rc **rc;
{
    if (PARSED_RC_magic != (*rc)->magic)
	panic("RC PANIC",__FILE__,__LINE__,"free_parsed_rc",
	      "Bad magic number",0);

    if ((*rc)->head_comments) {
	int i;

	for (i = 0; i < (*rc)->head_comments_len; i++) {
	    if ((*rc)->head_comments[i].buffer) {
		free((*rc)->head_comments[i].buffer);
		(*rc)->head_comments[i].buffer = NULL;
	    }
	    (*rc)->head_comments[i].buffer_len = 0;		
	}

	free((*rc)->head_comments);
	(*rc)->head_comments = NULL;
    }
    (*rc)->head_comments_len = 0;

    if ((*rc)->setting_lines) {
	int i;

	for (i = 0; i < (*rc)->setting_lines_len; i++) {

	    if ((*rc)->setting_lines[i].current_tag) {
		free((*rc)->setting_lines[i].current_tag);
		(*rc)->setting_lines[i].current_tag = NULL;
	    }

	    if ((*rc)->setting_lines[i].rest) {
		free((*rc)->setting_lines[i].rest);
		(*rc)->setting_lines[i].rest = NULL;
	    }
	    (*rc)->setting_lines[i].restlen = 0;

	}
	
	free((*rc)->setting_lines);
	(*rc)->setting_lines = NULL;
    }
    (*rc)->setting_lines_len = 0;

    (*rc)->magic = 0;   /* Invalidate */
    free(*rc);
    *rc = NULL;
}


static struct string * ascii_option_value P_((char *rest, int restlen,
					      int lineno, const char *filename));
static struct string * ascii_option_value(rest,restlen,lineno,filename)
     char *rest; 
     int restlen;
     int lineno; 
     const char *filename;
{
    struct string * rest_S = new_string(ASCII_SET);
    const char * csn  = get_charset_MIME_name(ASCII_SET);
    int ERRORS;
    int rs =  add_streambytes_to_string(rest_S,restlen,s2us(rest),&ERRORS);
    
    if (rs < restlen) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailedAsCharset,
			  "Failed to parse line %d as charset %s in \"%s\" file"),
		  lineno,
		  csn ? csn : "<no MIME name>",filename);

	free_string(&rest_S);

	return NULL;
    }

    if (ERRORS) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFileAsCharsetErrors,
			  "There is %d errors when parsing line %d as charset %s in \"%s\" file"),
		  ERRORS,lineno, csn ? csn : "<no MIME name>",filename);

	free_string(&rest_S);
	
	return NULL;	
    }

    return rest_S;
}

struct parsed_rc * do_rc_parse(file,lcl,filename,errors_p)
     FILE *file;
     enum record_mode lcl;
     const char *filename;
     int  * errors_p;
{
    struct parsed_rc *rc = NULL;
    struct rc_save_info_rec * prev_type = NULL;
    struct hist_record      * rec       = NULL;

    char current_tag[SLEN];
    int lineno = 0;

    int ignored = 0;
       
    char * buffer = NULL;
    int r;  /* length of line or error/eof status */

    int errors = *errors_p;

    if (!file) 
	return NULL;
    
    rc =  safe_malloc(sizeof (*rc));

    bzero((void *)rc, sizeof (*rc));
    rc->magic = PARSED_RC_magic;

    rc->head_comments     = NULL;
    rc->head_comments_len = 0;

    rc->setting_lines     = NULL;
    rc->setting_lines_len = 0;

    current_tag[0] = '\0';

    while (0 <= ( r = malloc_gets(&buffer,MAX_LINE,file))) {
	
	char   * tag         = NULL;
	char   keyword[SLEN];
	char   * rest        = NULL;
	int      restlen     = 0;

	int    assig  = 0;

	lineno++;

	if (!buffer)           /* EOF           */
	    break;
	if (r < 1)             /* empty line    */
	    continue;

	if (buffer[0] == '#') {   /* comment       */

	    if (lineno < 3) {
		
		rc->head_comments = 
		    safe_realloc(rc->head_comments,
				 (rc->head_comments_len +1) *
				 sizeof (rc->head_comments[0]));

		rc->head_comments[rc->head_comments_len].buffer
		    = safe_malloc(r+1);
		memcpy(rc->head_comments[rc->head_comments_len].buffer,
		       buffer,r+1);
		rc->head_comments[rc->head_comments_len].buffer_len = r;
		rc->head_comments[rc->head_comments_len].lineno = lineno;
		
		rc->head_comments_len++;		    
	    }

	    continue;
	}

	rest = rc_split_keyword(buffer,r,keyword, sizeof keyword,&tag,
				&assig, &restlen);

	if (tag && !keyword[0]) {
	    strfcpy(current_tag, tag, sizeof current_tag);
	    
	    if (rest[0] && '#' != rest[0]) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmIgnoredLine,
				  "Rest of line ignored after \"%s:\" on line %d on file %s"),
			  tag, lineno,filename);
		errors++;
	    }

	} else if (!keyword[0]) {
	    struct string * rest_S UNUSED_VAROK = NULL;

	    if (prev_type != NULL) {
		
		if (RCTYPE_magic != prev_type->dt_type->magic)
		    panic("RC PANIC",__FILE__,__LINE__,"do_rc_parse",
			  "Bad prev_type",0);

		if (rec) 
		    hist_record_add_line(rec,rest);

		DPRINT(Debug,10,(&Debug,"%s:%d: [cont] %s=%S\n",
				 filename,lineno,prev_type->name,rest_S));


		if (0 != (prev_type->flags & FL_ASCIIFIRST)) {
		    struct string * rest_S = ascii_option_value(rest,restlen,
								lineno,filename);

		    if (! rest_S) 
			goto have_errors_cline;
		    
		    if (!prev_type->dt_type -> 
			parse_cline(prev_type,
				    lcl,
				    rest_S,
				    lineno,
				    filename)) {

		    have_errors_cline:

			errors++;

			/* If there was errors, then assume that rewriting
			   of elmrc entry causes it to be rewritten
			*/
			if (rec)
			    hist_record_changed(rec);
		    }

		    if (rest_S)
			free_string(&rest_S);

		} else {

		    rc->setting_lines = 
			safe_realloc(rc->setting_lines,
				     (rc->setting_lines_len +1) * 
				     sizeof(rc->setting_lines[0]));

		    if (current_tag[0])
			rc->setting_lines[rc->setting_lines_len].current_tag
			    = safe_strdup(current_tag);
		    else
			rc->setting_lines[rc->setting_lines_len].current_tag = NULL;
		    
		    rc->setting_lines[rc->setting_lines_len].rest =
			safe_malloc(restlen+1);
		    memcpy(rc->setting_lines[rc->setting_lines_len].rest,rest,restlen+1);
		    rc->setting_lines[rc->setting_lines_len].restlen = restlen;
		    rc->setting_lines[rc->setting_lines_len].lineno = lineno;
		    rc->setting_lines[rc->setting_lines_len].is_cline = 1;
		    rc->setting_lines[rc->setting_lines_len].negate   = 0;
		    
		    rc->setting_lines[rc->setting_lines_len].record = prev_type;
		    rc->setting_lines[rc->setting_lines_len].hist_record = rec;
		    rc->setting_lines_len++;

		}
	    }

	} else if (assig) {  
	    int negate = 0;
	    
	    if (tag || current_tag[0]) {
#ifdef USE_DLOPEN
		struct rc_save_info_rec * rc_options      = NULL;
		int                     rc_option_count   = 0;
#endif		
		if (!tag) 
		    tag = current_tag;
		else
		    ignored = 0;
		
#ifdef USE_DLOPEN
		if (give_options(tag,&rc_options,&rc_option_count)) {
		    
		    prev_type = rc_locate_option(rc_options,rc_option_count,
						 keyword,&negate);
		    ignored = 0;
		    
		    if (!prev_type) {
			errors++;
			lib_error(CATGETS(elm_msg_cat, ElmSet, ElmBadKeyTagInElmrc,
					  "I can't understand keyword \"%s\" (section %s) in line %d in \"%s\" file"),
				  keyword, tag,lineno,filename);
			continue;
		    }
		    
		    
		} else if (!ignored) {
		    ignored = 1;
		    
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmIgnoredSection,
				      "Keywords on section \"%s\" ignored starting from line %d on file %s"),
			      tag, lineno,filename);
		    
		}
#else
		if (!ignored) {
		    ignored = 1;
		    
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmSharedIgnoredSection,
				      "Shared libraries not supported. Section \"%s\" ignored starting from line %d on file %s"),
			      tag, lineno,filename);
		    errors++;		
		}
#endif
		
	    } else {
		
		prev_type = rc_locate_option(save_info,NUMBER_OF_SAVEABLE_OPTIONS,
					     keyword,&negate);
		
		if (!prev_type) {
		    errors++;
		    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmBadKeyInElmrc,
				      "I can't understand keyword \"%s\" in line %d in \"%s\" file"),
			      keyword, lineno,filename);
		    
		    continue;
		}
		
	    }
	    
	    if (!prev_type) 
		continue;
	    
	    rec = rc_need_hist_record(lcl,prev_type,tag);
	    if (rec) 
		hist_record_first_line(rec,rest);
	    
	    if (ignored)
		continue;
	    
	    if (prev_type->flags & FL_SYS && lcl == LOCAL_RC) {
		lib_error(CATGETS(elm_msg_cat, ElmSet, 
				  ElmBadLocalKeyInElmrc,
				  "Keyword \"%s\" is not valid on user file %s"),
			  prev_type->name,filename);
		continue;
	    }
	    	    
	    if (0 != (prev_type->flags & FL_ASCIIFIRST)) {
		struct string * rest_S = ascii_option_value(rest,restlen,
							    lineno,filename);
		
		if (! rest_S) 
		    goto have_errors_line;
		
		if (!rc_eval_tail(prev_type,lcl,rest_S,lineno,filename,negate)) {
		    
		have_errors_line:
		    errors++;
		    
		    /* If there was errors, then assume that rewriting
		       of elmrc entry causes it to be rewritten
		    */
		    if (rec)
			hist_record_changed(rec);
		    
		}
		
		if (rest_S)
		    free_string(&rest_S);
		
	    } else {
		
		rc->setting_lines = 
		    safe_realloc(rc->setting_lines,
				 (rc->setting_lines_len +1) * 
				 sizeof(rc->setting_lines[0]));
		
		if (current_tag[0])
		    rc->setting_lines[rc->setting_lines_len].current_tag
			= safe_strdup(current_tag);
		else
		    rc->setting_lines[rc->setting_lines_len].current_tag = NULL;
		
		rc->setting_lines[rc->setting_lines_len].rest =
		    safe_malloc(restlen+1);
		memcpy(rc->setting_lines[rc->setting_lines_len].rest,rest,restlen+1);
		rc->setting_lines[rc->setting_lines_len].restlen  = restlen;
		rc->setting_lines[rc->setting_lines_len].lineno   = lineno;
		rc->setting_lines[rc->setting_lines_len].is_cline = 0;
		rc->setting_lines[rc->setting_lines_len].negate   = 
		    negate ? 1 : 0;
		
		rc->setting_lines[rc->setting_lines_len].record = prev_type;
		rc->setting_lines[rc->setting_lines_len].hist_record = rec;
		rc->setting_lines_len++;
	    }
	    
	} else {
	    prev_type = NULL;
	    
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmBadLineInElmrc,
			      "Keyword \"%s\" without = -character in line %d in \"%s\" file"),
		      keyword, lineno,filename);
	    errors++;
	}
    }
    
    if (-1 == r) {	
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmTooLongLine,
			  "%.30s: %d: Too long line: %.30s..."),
		  filename,lineno+1,
		  buffer ? buffer : "<not available>");
	errors++;
    }
    
    if (ferror(file)) {
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmErrorReading,
			  "Error reading from %.50s"),
		  filename);
	errors++;
    }
    
    if (buffer)
	free(buffer);
    
    *errors_p = errors;
    
    return rc;
}

struct editor_propline * current_editor_propline;

void do_rc_process(rc,lcl,filename,elmrc_charset, errors_p)
     struct parsed_rc *rc; 
     enum record_mode lcl;
     const char *filename; 
     charset_t elmrc_charset; /* NULL, if no @charset tag */
     int *errors_p;
{
    int i;

    charset_t rcset = system_charset;
    const char * csn;

    int errors = *errors_p;

     if (PARSED_RC_magic != rc->magic)
	panic("RC PANIC",__FILE__,__LINE__,"do_rc_process",
	      "Bad magic number",0);

     /* Nothing detected yet */
     if (current_editor_propline)
	 free_editor_propline(& current_editor_propline);
   
     if (editor_propline_ignore != editor_elmrc_propline.val) {
	 for (i = 0; 
	      i < rc->head_comments_len &&
		  !current_editor_propline; 
	      i++) {

	     current_editor_propline =
		 detect_editor_propline(editor_elmrc_propline.val,
					rc->head_comments[i].buffer,
					rc->head_comments[i].buffer_len,
					filename,
					rc->head_comments[i].lineno,&rcset);

	 }
     }

     if (elmrc_charset) {
	 csn  = get_charset_MIME_name(elmrc_charset);

	 rcset = elmrc_charset;
	 
	 DPRINT(Debug,20,(&Debug,
			  "do_rc_process: %s: Given %s charset\n",
			  filename,csn ? csn : "<no MIME name>"));
     } else {
	 csn  = get_charset_MIME_name(rcset);

	 DPRINT(Debug,20,(&Debug,
			  "do_rc_process: %s: Using %s as charset\n",
			  filename,csn ? csn : "<no MIME name>"));
     }


     for (i = 0; i < rc->setting_lines_len; i++) {
	 struct string * rest_S = new_string(rcset);
	 int code;
	 int ERRORS;
	 int rs =  
	     add_streambytes_to_string(rest_S,
				       rc->setting_lines[i].restlen,
				       cs2us(rc->setting_lines[i].rest),
				       &ERRORS);

	 if (rs < rc->setting_lines[i].restlen) {
	     lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFailedAsCharset,
			      "Failed to parse line %d as charset %s in \"%s\" file"),
		       rc->setting_lines[i].lineno,
		       csn ? csn : "<no MIME name>",filename);
	    errors++;
	    
	    goto fail;
	 }

	 if (ERRORS) {
	     lib_error(CATGETS(elm_msg_cat, ElmSet, ElmFileAsCharsetErrors,
			       "There is %d errors when parsing line %d as charset %s in \"%s\" file"),
		       ERRORS,rc->setting_lines[i].lineno,
		       csn ? csn : "<no MIME name>",filename);

	     errors++;
	 }

	 if (RCTYPE_magic != rc->setting_lines[i].record -> dt_type->magic)
	     panic("RC PANIC",__FILE__,__LINE__,"do_rc_process",
		   "Bad magic number for record",0);

	 if (rc->setting_lines[i].is_cline) 
	     code = rc->setting_lines[i].record -> dt_type->
		 parse_cline(rc->setting_lines[i].record,
			     lcl,rest_S,
			     rc->setting_lines[i].lineno,
			     filename);
	 else
	     code = rc_eval_tail(rc->setting_lines[i].record,
				 lcl,rest_S,
				 rc->setting_lines[i].lineno,
				 filename,
				 rc->setting_lines[i].negate);
				 
	 if (!code) {
	     errors++;
	     
	     /* If there was errors, then assume that rewriting
		of elmrc entry causes it to be rewritten
	     */
	     if (rc->setting_lines[i].hist_record)
		 hist_record_changed(rc->setting_lines[i].hist_record);
	 }
				 
     fail:
	 free_string(&rest_S);
     }

     *errors_p = errors;
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
