static char rcsid[] = "@(#)$Id: remote_mbx.c,v 2.23 2016/02/27 11:18:47 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.23 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                  (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "elm_defs.h"
#include "connection_imp.h"
#include "ss_imp.h"
#ifdef USE_DLOPEN
#include "shared_imp.h"
#endif
#include "s_me.h"
#include "s_elm.h"

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}

DEBUG_VAR(Debug,__FILE__,"net");

#ifdef POLL_METHOD
void zero_Read_Buffer(buffer)
     struct Read_Buffer *buffer;
{
    buffer->read_buffer = NULL;
    buffer->read_len    = 0;
}

void free_Read_Buffer(buffer)
     struct Read_Buffer *buffer;
{
    if (buffer->read_buffer) {
	free(buffer->read_buffer);
	buffer->read_buffer = NULL;
    }
    buffer->read_len = 0;
}

/* For non-lieral we read with quite short block because
   many answers are quite small
*/

#define READ_BLOCK   1024

int ReadFromSocket(fd,buffer,wanted)
     int fd; 
     struct Read_Buffer *buffer;
     int wanted;
{
    int n;

    if (wanted > 0) {
	buffer -> read_buffer = safe_realloc(buffer -> read_buffer,
					     buffer -> read_len + wanted);
	n = read(fd, buffer -> read_buffer + buffer -> read_len, wanted);
    } else {
	buffer -> read_buffer = safe_realloc(buffer -> read_buffer,
					     buffer -> read_len + READ_BLOCK);
	n = read(fd, buffer -> read_buffer + buffer -> read_len, READ_BLOCK);
    }

    return n;
}

int find_crlf(buffer, add_null)
     struct Read_Buffer *buffer;
     int add_null;
{
    char * p = buffer->read_buffer;
    int i;

    for (i = 0; i < buffer->read_len -1; i++)
	if ('\r' == p[i] && '\n' == p[i+1]) {
	    if (add_null) {
		p[i] = '\0';
		p[i+1] = '\0';
	    }
	    DPRINT(Debug,49,(&Debug,
			     "find_crlf=%d (%d total)\n",
			     i+2,buffer->read_len));
	    return i+2;
	}
    DPRINT(Debug,49,(&Debug,
		     "find_crlf=0 (%d total)\n",
		     buffer->read_len));
    return 0;
}

void cut_line(buffer, len)
     struct Read_Buffer *buffer;
     int len;
{
    char * p = buffer->read_buffer;
    
    buffer->read_len -= len;
    DPRINT(Debug,49,(&Debug,
		     "cut_line: %d chars consumed, %d left\n",
		     len,buffer->read_len));
    if (buffer->read_len)
	memmove(p, p+len, buffer->read_len);
}     

void zero_Write_Buffer(buffer)
     struct Write_Buffer *buffer;
{
    buffer->write_buffer = NULL;
    buffer->write_len    = 0;
}

void free_Write_Buffer(buffer)
     struct Write_Buffer *buffer;
{
    if (buffer->write_buffer) {
	free(buffer->write_buffer);
	buffer->write_buffer = NULL;
    }
    buffer->write_len = 0;
}

int WriteToSocket(fd,buffer)
     int fd; 
     struct Write_Buffer *buffer;
{
    int n = write(fd,buffer->write_buffer, buffer->write_len);
    
    return n;
}

void cut_Write_Buffer(buffer,n)
     struct Write_Buffer *buffer; 
     int n;
{
    if (n > 0) {
	char * p = buffer->write_buffer;
	
	buffer->write_len -= n;
	if (buffer->write_len)
	    memmove(p, p+n, buffer->write_len);

	DPRINT(Debug,13,(&Debug,
			 "cut_Write_Buffer: Written %d bytes (%d left)\n",
			 n,buffer->write_len));
    }
}

void add_to_Write_Buffer(buffer,str,l)
     struct Write_Buffer *buffer;
     char **str;
     int l;
{
    if (!buffer->write_len) {
	if (buffer->write_buffer) 
	    free(buffer->write_buffer);
	buffer->write_buffer = *str;
	buffer->write_len    = l;
    } else if (l > 0) {
	buffer->write_buffer = safe_realloc(buffer->write_buffer,
					    buffer->write_len+l);
	memmove(buffer->write_buffer + buffer->write_len,*str,l);
	buffer->write_len += l;
	free(*str);
    }
    *str = NULL;
}

#endif

/* Seems that h_errno is macro on AIX */
#ifndef h_errno
extern int h_errno;
#endif

#include <errno.h>
#ifndef ANSI_C
extern int errno;
#endif

/* Closed only implicity on exit */
FILE * transaction_file = NULL;

int set_transaction_file(filename)
     const char *filename;
{
    FILE * f;
    int err;
    int fd;


    if (userid == -1 || groupid == -1) {
	/* save original user and group ids */
	userid  = getuid();
	groupid = getgid();	

	DPRINT(Debug,1,(&Debug,"set_transaction_file: %s: Saving userid/groupid=%d/%d\n",
			filename,userid,groupid));

    }

    err = can_open(filename,"a");
    if (0 != err) {
	lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotWriteable,
			  "File %.50s is not writeable: %s"),
		  filename, strerror(err));
	return 0;
    }

    fd = open(filename,O_CREAT|O_WRONLY|O_APPEND,00600);
    if (-1 == fd) {
	int err = errno;
	lib_error(CATGETS(elm_msg_cat, MeSet, MeFileNotWriteable,
			  "File %.50s is not writeable: %s"),
		  filename, strerror(err));
	return 0;
    }
    elm_chown(filename, userid, groupid); /* file owned by user */


    f = fdopen(fd,"a");
    if (!f) {
	close(fd);
	return 0;
    }

    if (transaction_file) {
	fprintf(transaction_file,
		"\n===== Changing logging to file %s\n",
		filename);
	fclose(transaction_file);
    }
    transaction_file = f;
    
#ifdef SETLINEBUF
    setlinebuf(transaction_file);
#endif

    DPRINT(Debug,1,(&Debug,
		    "set_transaction_file: %s: transaction file opened\n",
		    filename));

    return 1;
}

#ifdef REMOTE_MBX


void zero_remote_account(ra)
     struct remote_account *ra;
{
    /* bzero is defined hdrs/elm_defs.h */
    bzero ((void *)ra, sizeof (struct remote_account));

    ra->magic                 = REMOTE_ACCOUNT_magic;

    ra->addrsize              = 0;
    ra->hostaddr.dummy        = NULL;
    ra->hostname              = NULL;
    ra->service               = NULL;
    ra->stream                = NULL;
    ra->username              = NULL;
    ra->host                  = NULL;
}

int remote_account_OK(st_size)
     size_t st_size;
{

    if (st_size == sizeof(struct remote_account))
	return 1;

    DPRINT(Debug,1,(&Debug,
		    "remote_account_OK: Size mismatch st_size=%d != %d\n",
		    st_size,sizeof(struct remote_account)));

    return 0;
}

void clear_remote_account(ra)
     struct remote_account *ra;
{
    if (ra->magic  != REMOTE_ACCOUNT_magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "clear_remote_account",
	      "Bad magic number",0);

    if (ra->hostaddr.dummy) {
	free(ra->hostaddr.dummy);
	ra->hostaddr.dummy = NULL;
    }
    ra->addrsize      = 0;

    if (ra->hostname) {
	free(ra->hostname);
	ra->hostname = NULL;
    }

    if (ra->service)
	free_service_entry(& (ra->service));

    if (ra->stream) 
	FreeStreamStack(& (ra->stream));
    if (ra->username) {
	free(ra->username);
	ra->username = NULL;
    }
    if (ra->host) {
	free(ra->host);
	ra->host = NULL;
    }
}

#if defined(I_NETINET_IN) && defined(I_ARPA_INET)



static int connect_one_IN P_((struct remote_account *ra,
			      int *cur_socket, int *last_error,
			      struct cancel_data *can,
			      int report_port));
static int connect_one_IN(ra,cur_socket,last_error,can,
			  report_port)
     struct remote_account *ra;
     int *cur_socket;
     int *last_error;
     struct cancel_data *can;
     int report_port;
{
    int r = 0;

    char mybuffer[256];
    const char * s;

    if (ra->magic  != REMOTE_ACCOUNT_magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "connect_one_IN",
	      "Bad magic number",0);
    
    if (!ra->hostaddr.dummy) {
	DPRINT(Debug,12,(&Debug,
			 "connect_one_IN: No address\n"));
	return -1;
    }

    s = give_addr_as_string(ra->hostaddr.generic,mybuffer,sizeof mybuffer);
			    
    DPRINT(Debug,12,(&Debug,
		     "connect_one_IN: family=%d size=%d",
		     ra->hostaddr.sa->sa_family,
		     ra->addrsize));
    
    if (s) {
	DPRINT(Debug,12,(&Debug," addr=%s",s));
    }
    if (ra->hostname) {
	DPRINT(Debug,12,(&Debug," hostname=%s",ra->hostname));
    }

    switch (ra->hostaddr.sa->sa_family) {
    case AF_INET:
	DPRINT(Debug,12,(&Debug," IPv4 port=%d\n",
			 ntohs(ra->hostaddr.sin->sin_port)));
	
	if (report_port)
	    cancel_progress(can,CATGETS(elm_msg_cat, MeSet,MeConPort,
					"Trying port %d"),
			    ntohs(ra->hostaddr.sin->sin_port));

	break;
#ifdef HAVE_IN6
    case AF_INET6: 
	DPRINT(Debug,12,(&Debug," IPv6 port=%d\n",
			 ntohs(ra->hostaddr.sin6->sin6_port)));


	if (report_port)
	    cancel_progress(can,CATGETS(elm_msg_cat, MeSet,MeConPort,
					"Trying port %d"),
			    ntohs(ra->hostaddr.sin6->sin6_port));
	
	break;
#endif
    default:
	DPRINT(Debug,12,(&Debug," ????\n"));
	break;
    }
    
    /* r == 0: not succeed
       r <  0: fatal error
       r >  0: succeed
    */


 retry:
    if (-1 != (*cur_socket)) {
	/* Some systems require that socket is closed even
	   after unsuccesfully connect() -- it can not be reused
	*/
	DPRINT(Debug,12,(&Debug,
			 "connect_one_IN:  closing socket (%d) for reopening\n",
			 (*cur_socket)));;
	close(*cur_socket);
	(*cur_socket) = -1;
    }
    
    if (-1 == (*cur_socket)) {
	(*cur_socket) = socket(ra->hostaddr.sa->sa_family,
			       SOCK_STREAM,IPPROTO_TCP);
	if (-1 == (*cur_socket)) {
	    int err =errno;
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeFailedToCreate,
			      "Failed to create socket: %s"),
		      strerror(err));
	    r = -1;
	    goto clean;
	}
	DPRINT(Debug,12,(&Debug,
			 "connect_one_IN: socket=%d\n",(*cur_socket)));
    }
    	
    if (-1 == connect((*cur_socket),ra->hostaddr.sa,
		      ra->addrsize)) {
	(*last_error) = errno;
	DPRINT(Debug,12,(&Debug,
		    "connect_one_IN: connect failed: %s\n",
		    strerror(*last_error)));
	
	if (0 == r && EINTR == *last_error) {
	    
	    if (is_canceled(can)) {
		DPRINT(Debug,12,(&Debug,"connect canceled...\n"));
		r = -1;
	    } else {
		DPRINT(Debug,12,(&Debug,"Retrying connect...\n"));
		goto retry;
	    }
	    
	}
    } else
	r = 1;
		
    /* If got connection to host and and it refused
       connection, try another ports
    */
    if (r < 1 && *last_error != ECONNREFUSED && *last_error != ETIMEDOUT) {
	DPRINT(Debug,12,(&Debug,
			 "connect_one_IN: Treating as fatal error\n"));
	r = -2;			   
    }

 clean:
    DPRINT(Debug,12,(&Debug,
		     "connect_one_IN=%d%s\n",r,r > 0 ? " (succeed)" : ""));
    return r;
}
#endif

/* Return 1 if succeed, 0 if too small*/
static int set_SOCKADDR_ptr_helper P_((size_t                 * target_addrsize,
				       union SOCKADDR_ptr     * target,
				       const size_t             source_addrsize,
				       const union SOCKADDR_ptr source));

int connect_remote_account_1(ra,ra_from)
     struct remote_account *ra;  
     const struct remote_account *ra_from;
{
    int cur_socket   = -1;
    int last_error   = 0;
    int ok = 0;
    int was_canceled = 0;
    int port = 0;

    if (ra->magic  != REMOTE_ACCOUNT_magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "connect_remote_account_1",
	      "Bad magic number",0);
    
    if (ra->stream) {
	DPRINT(Debug,12,(&Debug,
			 "connect_remote_account_1: Closing for reopening\n"));
	FreeStreamStack(& (ra->stream));
    }

    if (ra->service) {
	DPRINT(Debug,12,(&Debug,
			 "connect_remote_account_1: Removing old service entry: %p\n",
			 ra->service));	 
	free_service_entry(& (ra->service));
    }

    if (ra_from->magic  != REMOTE_ACCOUNT_magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "connect_remote_account_1",
	      "Bad (ra_from) magic number",0);
       
    if (! ra_from->hostaddr.dummy ||
	! ra_from->addrsize        ||
	! ra->username             ||
	! ra->host)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "connect_remote_account_1",
	      "No data (ra_from)",0);

    if (! set_SOCKADDR_ptr_helper(& (ra->addrsize),
				  & (ra->hostaddr),
				  ra_from->addrsize,
				  ra_from->hostaddr)) {
	ok = 0;
	goto clean;
    }

    /* name corresponding to hostaddr */
    ra->hostname = NULL;
    if (ra_from->hostname)
	ra->hostname = strmcpy(ra->hostname,
			       ra_from->hostname);

    /* Copy username@host */
    ra->username = strmcpy(ra->username, ra_from->username);
    ra->host     = strmcpy(ra->host, ra_from->host);

    switch (ra->hostaddr.sa->sa_family) {
#if defined(I_NETINET_IN) && defined(I_ARPA_INET)
    case AF_INET:        
	port = ntohs(ra->hostaddr.sin->sin_port);
	ok = 1;
	break;

#ifdef HAVE_IN6
    case AF_INET6: 	
	port = ntohs(ra->hostaddr.sin6->sin6_port);
	ok = 1;    
	break;
#endif

#endif
    default:
	lib_error(CATGETS(elm_msg_cat, MeSet,MeUnsupportedAddrType,
			  "Name %s have odd type address"),
		  ra->hostname ? ra->hostname : ra->host);
	break;
    }

    if (ok) {
	struct cancel_data * can = NULL;
	int r;
	char mybuffer[256];

	const char * s = give_addr_as_string(ra->hostaddr.generic,
					     mybuffer,sizeof mybuffer);
	
	if (ra->hostname && 0 != strcmp(ra->host,ra->hostname))
	    can = new_cancel(CATGETS(elm_msg_cat, MeSet,MeConnectingPort2,
				 "Connecting to %s [%s/%s], port %d ..."),
			     ra->host,
			     ra->hostname,
			     s ? s : "???",
			     port);
	else
	    can = new_cancel(CATGETS(elm_msg_cat, MeSet,MeConnectingPort1,
				     "Connecting to %s [%s], port %d ..."),
			     ra->host,
			     s ? s : "???",
			     port);
	
	r = connect_one_IN(ra,&cur_socket,&last_error,can,0);

	if (r > 0) {
	    
	    DPRINT(Debug,12,(&Debug,
			     "connect_remote_account_1: Connection succeed %s [%s/%s], port %d\n",
			     ra->host,
			     ra->hostname ? ra->hostname : "???",
			     s ? s : "???",
			     port));
	    ok = 1;
	    
	    if (transaction_file) {
		time_t tmp = time(NULL);
		struct tm * zz = localtime(&tmp);
		fprintf(transaction_file,
			"%d [%d] %02d:%02d:%02d === CONNECT %s [%s/%s], port %d\n",
			getpid(),cur_socket,
			zz ? zz->tm_hour : 00,
			zz ? zz->tm_min  : 00,
			zz ? zz->tm_sec  : 00,
			ra->host,
			ra->hostname ? ra->hostname : "???",
			s ? s : "???",
			port);
	    }

	} else
	    ok = 0;

	was_canceled = is_canceled(can);
	free_cancel(&can);
    }

    if (!ok) {

	if (cur_socket != -1) {
	    DPRINT(Debug,12,(&Debug,
			     "connect_remote_account_1: Closing socket (%d) after failure\n",
			     cur_socket));
	    close(cur_socket);
	    cur_socket = -1;	    
	}

	if (was_canceled)
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeConnectCanceled,
			      "Connect %s canceled."),
		      ra->host);
	else if (last_error)
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeConnectFailed,
			      "Failed to connect %s: %s"),
		      ra->host,strerror(last_error));
        else
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeConnectFailedX,
			      "Failed to connect %s"),
		      ra->host);

	ok = 0;
	goto clean;
    }
    
    /* We set folder to non-blocking after connect, because
     * non-blocking connect is little complicate to use
     */
    
    if (-1 == fcntl(cur_socket,F_SETFL,O_NONBLOCK)) {
	int err UNUSED_VAROK = errno;
	DPRINT(Debug,12,(&Debug,
			 "connect_remote_account_1: fcntl [O_NONBLOCK] failed: %s\n",
			 strerror(err)));
    }

    ra->stream      = returnSimpleStream(cur_socket);
    ra->service     = ra_from->service;
    if (ra->service)
	inc_service_entry_refcount(ra->service);

 clean:
    DPRINT(Debug,12,(&Debug,
		     "connect_remote_account_1=%d\n",ok)); 
    return ok;
}


void set_address_from_gen_helper(addrsize,ptr,generic)
     size_t             * addrsize;
     union SOCKADDR_ptr * ptr;
     const SOCKADDR     * generic;
{

    switch (generic->sa.sa_family) {
    case AF_INET:
	ptr->sin    = safe_realloc(ptr->sin,
				   sizeof(generic->sin));
	*(ptr->sin) = generic->sin;
	*addrsize   = sizeof(generic->sin);
	break;
#ifdef HAVE_IN6
    case AF_INET6: 
	ptr->sin6  = safe_realloc(ptr->sin6,
				  sizeof(generic->sin6));
	*(ptr->sin6) = generic->sin6;
	*addrsize         =  sizeof(generic->sin6);
	break;
#endif
    default:
	DPRINT(Debug,11,(&Debug,
			 "set_address_from_gen_helper: Unknown address family %d\n",
			 generic->sa.sa_family));

	ptr->generic    = safe_realloc(ptr->generic,
				       sizeof (*generic));
	*(ptr->generic) = *generic;
	*addrsize       =  sizeof(*generic);
	break;
    }
}


/* Return 1 if succeed, 0 if too small*/
static int set_SOCKADDR_ptr_helper(target_addrsize,target,
				   source_addrsize,source)
     size_t                 * target_addrsize;
     union SOCKADDR_ptr     * target;
     const size_t             source_addrsize;
     const union SOCKADDR_ptr source;
{
    /* sa_family is assumed to be first field !! */
    if (source_addrsize < sizeof (source.sa->sa_family)) {
	DPRINT(Debug,11,(&Debug,
			 "set_SOCKADDR_ptr_helper=0; size=%d\n",
			 source_addrsize));
	goto fail;
    }

    switch (source.sa->sa_family) {
    case AF_INET:
	if (source_addrsize < sizeof (* (source.sin))) {
	    DPRINT(Debug,11,(&Debug,
			     "set_SOCKADDR_ptr_helper=0; size=%d\n",
			     source_addrsize));
	    goto fail;
	}

	target->sin      = safe_realloc(target->sin,
					sizeof(* (source.sin)));
	*(target->sin)   = *(source.sin);
	*target_addrsize = sizeof(* (source.sin));
	break;
#ifdef HAVE_IN6
    case AF_INET6: 
	if (source_addrsize < sizeof (* (source.sin6))) {
	    DPRINT(Debug,11,(&Debug,
			     "set_SOCKADDR_ptr_helper=0; size=%d\n",
			     source_addrsize));
	    goto fail;
	}
	target->sin6  = safe_realloc(target->sin6,
				     sizeof(* (source.sin6)));
	*(target->sin6) = * (source.sin6);
	*target_addrsize  =  sizeof(* (source.sin6));
	break;
#endif
    default:
	DPRINT(Debug,11,(&Debug,
			 "set_SOCKADDR_ptr_helper: Unknown address family %d, size %d\n",
			 source.sa->sa_family,source_addrsize));
	
	/* Copy all information */
	target->dummy = safe_realloc(target->dummy,source_addrsize);	    
	memcpy(target->dummy,source.dummy,source_addrsize);
	*target_addrsize = source_addrsize;
	break;
    }

    return 1;

 fail:   /* Clear target on failure */
    *target_addrsize = 0;
    if (target->dummy) {
	free(target->dummy);
	target->dummy = NULL;
    }
    return 0;
}
					    

/* Return 1 if succeed, 0 if too small */
static int ra_set_address_from_se_address P_((struct remote_account *ra,
					      struct service_entry_addr *ad,
					      const char * official_name));
static int ra_set_address_from_se_address(ra,ad,official_name)
     struct remote_account *ra;
     struct service_entry_addr *ad;
     const char * official_name;
{
    int ret ;

    if (ra->magic  != REMOTE_ACCOUNT_magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "ra_set_address_from_se_address",
	      "Bad magic number",0);
    
    if (ad->hostname)
	ra->hostname = strmcpy(ra->hostname,ad->hostname);
    else
	ra->hostname = strmcpy(ra->hostname,official_name);

    ret = set_SOCKADDR_ptr_helper(& (ra->addrsize),
				   & (ra->hostaddr),
				   ad->addrsize,
				   ad->address);

    if (!ret) {
	free(ra->hostname);
	ra->hostname = 0;
    }	

    return ret;
}

#if NOTUSED
static void ra_set_address_from_generic P_((struct remote_account *ra,
					   const SOCKADDR *generic));
static void ra_set_address_from_generic(ra,generic)
     struct remote_account *ra;
     const SOCKADDR *generic;
{
    if (ra->magic  != REMOTE_ACCOUNT_magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "ra_set_address_from_generic",
	      "Bad magic number",0);
 
    set_address_from_gen_helper(& ( ra->addrsize ),
				& ( ra->hostaddr ),
				generic);   
}
#endif

int connect_remote_account(ra,got,se,default_portlist,force_port)
     struct remote_account *ra;     
     int *got;
     struct service_entry *se;    
     PORTS default_portlist[];
     PORTS force_port;
{
    int ok = 0;
    int last_error   = 0;
    int cur_socket   = -1;
    int prev_family  = AF_UNSPEC;
    int was_canceled = 0;

    int idx;

    if (ra->magic  != REMOTE_ACCOUNT_magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "connect_remote_account",
	      "Bad magic number",0);

    if ( SERVICE_ENTRY_magic != se->magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "connect_remote_account",
	      "Bad magic (service_entry)",0);

    if(ra->stream) {
	DPRINT(Debug,12,(&Debug,
			 "connect_remote_account: Closing for reopening\n"));
	FreeStreamStack(& (ra->stream));
    }

    if (ra->service) {
	DPRINT(Debug,12,(&Debug,
			 "connect_remote_account: Removing old service entry: %p\n",
			 ra->service));	 
	free_service_entry(& (ra->service));
    }

    for (idx = 0; idx < se->addr_count; idx++) {

	/* Copy all information */
	
	if (!se->addr_list[idx].address.dummy) {
	    DPRINT(Debug,12,(&Debug,
			     "connect_remote_account: %d: No address\n",idx));
	    continue;
	}

	if (! ra_set_address_from_se_address(ra,& (se->addr_list[idx]),
					     se->official_name)) {
	    DPRINT(Debug,12,(&Debug,
			     "connect_remote_account: %d: bad address\n",idx));
	    continue;
	}					     

	if (! ra->hostaddr.sa)
	    panic("CONNECTION PANIC",__FILE__,__LINE__,
		  "connect_remote_account",
		  "No address",0);
       
	if (prev_family != ra->hostaddr.sa->sa_family &&
	    cur_socket != -1) {

	    DPRINT(Debug,12,(&Debug,
			     "connect_remote_account: %d: family change from %d to %d -- closing socket (%d)\n",
			     idx,prev_family,ra->hostaddr.sa->sa_family,
			     cur_socket));

	    close(cur_socket);
	    cur_socket = -1;	    
	}

	switch (ra->hostaddr.sa->sa_family) {
#if defined(I_NETINET_IN) && defined(I_ARPA_INET)
	case AF_INET: 
#ifdef AF_INET6
	case AF_INET6: 	
#endif
	    {
		struct cancel_data * can = NULL;
		char mybuffer[256];
		SOCKADDR A;
		const char *s; 
		int r = 0;
		int port;
		
		if (! set_SOCKADDR_from_ptr_helper(& A,
						   ra->addrsize,
						   ra->hostaddr)) {
		    DPRINT(Debug,11,(&Debug," --- bad address\n"));
		    continue;
		}

		s = give_addr_as_string(&A,mybuffer,sizeof mybuffer);
		port = give_addr_port(&A);

		/* r == 0: not succeed
		   r <  0: fatal error
		   r >  0: succeed
		*/

		if (force_port != PORT_end) {
		    
		    DPRINT(Debug,11,(&Debug,
				     "-- force_port = %d\n", force_port));
		    
		    if (PORT_end != port &&
			port != force_port)  {
			
			DPRINT(Debug,11,(&Debug,
					 "   but address gives port %d\n",
					 port));
			continue;
		    }
		    
		    if (ra->hostname && 
			0 != strcmp(se->official_name,ra->hostname))
			can = new_cancel(CATGETS(elm_msg_cat, MeSet,
						 MeConnectingPort3,
					     "Connecting to %s [%s/%s], port %d ... (%d)"),				     
					 se->official_name,
					 ra->hostname,
					 s ? s : "???",
					 force_port,
					 idx);
		    else
			can = new_cancel(CATGETS(elm_msg_cat, MeSet,
						 MeConnectingPort,
						 "Connecting to %s [%s], port %d ... (%d)"),				     
					 se->official_name,
					 s ? s : "???",
					 force_port,
					 idx);
		    
		    set_addr_port(ra->hostaddr.generic,force_port);
		    r = connect_one_IN(ra,&cur_socket,&last_error,can,1);
		    if (r > 0)
			*got = force_port;
		} else if (PORT_end != port) {
		    
		    DPRINT(Debug,11,(&Debug,"-- address record gives port %d\n",
				     port));
		    
		    if (ra->hostname && 
			0 != strcmp(se->official_name,ra->hostname))
			can = new_cancel(CATGETS(elm_msg_cat, MeSet,
						 MeConnectingPort3,
						 "Connecting to %s [%s/%s], port %d ... (%d)"),				     
					 se->official_name,
					 ra->hostname,
					 s ? s : "???",
					 port,
					 idx);
		    else
			can = new_cancel(CATGETS(elm_msg_cat, MeSet,
						 MeConnectingPort,
						 "Connecting to %s [%s], port %d ... (%d)"),
					 se->official_name,
					 s ? s : "???",
					 port,
					 idx);
		    
		    r = connect_one_IN(ra,&cur_socket,&last_error,can,0);
		    if (r > 0)
			*got = port;
		    
		} else {
		    
		    if (ra->hostname && 
			0 != strcmp(se->official_name,ra->hostname))
			can = new_cancel(CATGETS(elm_msg_cat, MeSet,
						 MeConnecting2,
						 "Connecting to %s [%s/%s]... (%d)"),
					 se->official_name,
					 ra->hostname,
					 s ? s : "???",
					 idx);
		    else
			can = new_cancel(CATGETS(elm_msg_cat, MeSet,
						 MeConnecting,
						 "Connecting to %s [%s]... (%d)"),
					 se->official_name,
					 s ? s : "???",
					 idx);
		    
		    if (se->port_count > 0) {
			
			int idx2;
			DPRINT(Debug,11,(&Debug,
					 "-- service list gives portlist (%d ports)\n",
					 se->port_count));
			
			for (idx2 = 0; idx2 < se->port_count && 0 == r; idx2++) {
			    
			    set_addr_port(ra->hostaddr.generic,
					  se->port_list[idx2]);
			    r = connect_one_IN(ra,&cur_socket,&last_error,can,1);
			    if (r > 0)
				*got = se->port_list[idx2];
			    
			}
			
			/* Do not use default portlist if port was 
			   given by name */
		    } else if (0 == (se->flags & SE_rescan_name_port)) {
			int idx2;
			
			
			DPRINT(Debug,11,(&Debug,
					 "-- service list does not give portlist -- trying default portlist\n"));
			
			for (idx2 = 0; default_portlist[idx2] != PORT_end && 0 == r; 
			     idx2++) {
			    
			    set_addr_port(ra->hostaddr.generic,
					  default_portlist[idx2]);
			    r = connect_one_IN(ra,&cur_socket,&last_error,can,1);
			    if (r > 0)
				*got = default_portlist[idx2];
			}
		    }
		}	
		
		if (r > 0) {
		    
		    DPRINT(Debug,12,(&Debug,
				     "connect_remote_account: Connection succeed %s [%s/%s], port %d\n",
				     se->official_name,
				     	ra->hostname ? ra->hostname : "???",
				     s ? s : "???",
				     *got));
		    ok = 1;
		    
		    if (transaction_file) {
			time_t tmp = time(NULL);
			struct tm * zz = localtime(&tmp);
			fprintf(transaction_file,
				"%d [%d] %02d:%02d:%02d === CONNECT %s [%s/%s], port %d\n",
				getpid(),cur_socket,
				zz ? zz->tm_hour : 00,
				zz ? zz->tm_min  : 00,
				zz ? zz->tm_sec  : 00,
				se->official_name,
				ra->hostname ? ra->hostname : "???",
				s ? s : "???",
				*got);
		    }
		}
		
		was_canceled = is_canceled(can);
		free_cancel(&can);
	    }	       
	    break;
#endif
	default:
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeUnsupportedAddrType,
			      "Name %s have odd type address"),
		       ra->hostname ? ra->hostname : se->official_name);
	    break;
	}
	
		
	if (ok) {
	    DPRINT(Debug,20,(&Debug,
			     "connect_remote_account: Connection succeed -- quiting loop\n"));
	    break;
	}
    }

    if (!ok) {

	if (cur_socket != -1) {
	    DPRINT(Debug,12,(&Debug,
			     "connect_remote_account: Closing socket (%d) after failure\n",
			     cur_socket));
	    close(cur_socket);
	    cur_socket = -1;	    
	}

	if (was_canceled)
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeConnectCanceled,
			      "Connect %s canceled."),
		      se->official_name);
	else if (last_error)
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeConnectFailed,
			      "Failed to connect %s: %s"),
		      se->official_name,strerror(last_error));
	else
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeConnectFailedX,
			      "Failed to connect %s"),
		      se->official_name);

	ok = 0;
	goto clean;
    }

    /* We set folder to non-blocking after connect, because
     * non-blocking connect is little complicate to use
     */
    
    if (-1 == fcntl(cur_socket,F_SETFL,O_NONBLOCK)) {
	int err UNUSED_VAROK = errno;
	DPRINT(Debug,12,(&Debug,
			 "connect_remote_account: fcntl [O_NONBLOCK] failed: %s\n",
			 strerror(err)));
    }

    ra->stream = returnSimpleStream(cur_socket);
    ra->service = se;
    inc_service_entry_refcount(ra->service);

 clean:
    DPRINT(Debug,12,(&Debug,
		     "connect_remote_account=%d\n",ok)); 
    return ok;
}

/* -1 == name not found or bad syntax
   0 == not a remote address
   1 == name found 
*/
int split_remote_name(name,X,se,rest,lookup_flags)
     const char *name;
     struct remote_account *X;
     struct service_entry **se;
     char **rest;
     int lookup_flags;
{
    char *sep;
    int ret = 0;
    
    DPRINT(Debug,12,(&Debug,
		     "split_remote_name: name=\"%s\", lookup_flags=%d\n",
		     name,lookup_flags));
    
    if (*se) {
	if (SERVICE_ENTRY_magic != (*se)->magic)
	    panic("CONNECTION PANIC",__FILE__,__LINE__,
		  "split_remote_name",
		  "Bad magic (service_entry)",0);

	free_service_entry(se);
    }

    *rest = NULL;
    zero_remote_account(X);
    
    *rest = strpbrk(name,"/:");
    sep  = strchr(name,'@');

    if (sep && (!*rest || *rest > sep)) {
	char * sep2;
       
	if (sep == name || sep[1] == '\0' ||
	    *rest == sep+1) {
	    lib_error(CATGETS(elm_msg_cat, MeSet,MeBadRemoteMailbox,
			      "Bad remote mailbox: %s"),
		      name);
	    ret = -1;
	    goto clean;
	}

	/* Test for '@' on hostname -- if there is, then
	   assume that first '@' was on username
	*/
	sep2 = strchr(sep+1,'@');
	if (sep2 && (!*rest || *rest > sep2+1)) {
	    DPRINT(Debug,12,(&Debug,
			     "split_remote_name: Found second '@' from hostname, Assuming that first '@' was on username\n"));
	    sep = sep2;
	}
	
	if (*rest) {
	    const char * START = sep+1;
	    X->host = safe_malloc(*rest - START+1);
	    memcpy(X->host,START,*rest - START);
	    X->host[*rest - START] = '\0';
	} else
	    X->host = safe_strdup(sep+1);	   

	X->username = safe_malloc(sep - name+1);
	memcpy(X->username,name,sep - name);
	X->username[sep - name] = '\0';

	if (0 == lookup_flags) {
	    lookup_flags = *rest ? STFLAG_browser : STFLAG_mbox;

	    DPRINT(Debug,12,(&Debug,
			     "split_remote_name: (%s) Using lookup_flags=%d\n",
			     name,lookup_flags));
	}

	*se = give_service_entry(X->host, lookup_flags);
	if (!*se) {
	    ret = -1;
	    goto clean;
	}

	/* Canonify host name ... */
	free(X->host);
	X->host = safe_strdup((*se)->official_name);

	DPRINT(Debug,12,(&Debug,
			 "split_remote_name: username=%s, host=%s\n",
			 X->username ? X->username : "<NULL>",
			 X->host ? X->host : "<NULL>"));

	ret = 1;
    } 
    
 clean:
    DPRINT(Debug,12,(&Debug,
		     "split_remote_name=%d: *rest=%s\n",
		     ret,
		     *rest ? *rest : "<NULL>"));

    return ret;
}		

/* 1 verify OK
   0 verify failed
  -1 no connection or other failure
*/

int remote_account_verify_peer(ra,data,cb,Server,server)
     struct remote_account        * ra;
     union ra_verify_peer_cb_data * data;
     ra_verify_peer_cb_f          * cb;
     const struct string          * Server;
     const struct string          * server;
{
    int ret = 1;

    int r;

    if (ra->magic  != REMOTE_ACCOUNT_magic)
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "remote_account_verify_peer",
	      "Bad magic number",0);

    if (! ra->stream) {
	DPRINT(Debug,8,(&Debug,
			"remote_account_verify_peer: No connection\n"));

	ret = -1;
	goto failure;
    }
    
    if (ra->service) {

	if (SERVICE_ENTRY_magic != ra->service->magic)
	    panic("CONNECTION PANIC",__FILE__,__LINE__,
		  "remote_account_verify_peer",
		  "Bad magic number (service_entry)",0);
    
	if (! (ra->service->flags & SE_ver_tls_cert) &&
	    ! (ra->service->flags & SE_require_tls_name) &&
	    ra->service->req_tls_peername_count < 1) {

	    DPRINT(Debug,8,(&Debug,
			    "remote_account_verify_peer: No verify required.\n"));

	    goto failure;
	}

	r = cb(ra,data,Server,server);

	if (r <= 0) {
	    DPRINT(Debug,8,(&Debug,"remote_account_verify_peer: callback=%d\n",
			    r));
	    ret = r;
	    goto failure;
	}

	if ((ra->service->flags & SE_ver_tls_cert)) {
	    r = 0;

	    StreamInfo(ra->stream,SS_verify,&r,NULL,NULL);

	    if (r < 0) {

		lib_error(CATGETS(elm_msg_cat, MeSet, MeRAFailedVerifyCert,
				  "Failed to verify %S %s certificate (required)."),
			  server,ra->host);
		ret = 0;
		goto failure;
	    } else if (0 == r) {
		int bits;
		
		StreamInfo(ra->stream,SS_ssf,&bits,NULL,NULL);
		
		if (bits > 0)
		    lib_error(CATGETS(elm_msg_cat, MeSet, 
				      MeRANotVerifiedCert,  
				      "%S %s certificate not verified (required)."),
			      Server,ra->host);
		else
		    lib_error(CATGETS(elm_msg_cat, MeSet, 
				      MeRANotEncryptedCert,
				      "%S %s connection not encrypted (certificate required)."),
			      Server,ra->host);
		ret = 0;
		goto failure;
	    }
	}

	if ((ra->service->flags & SE_require_tls_name) ||
	    ra->service->req_tls_peername_count > 0) {

	    int i;
	    int start = 0;

	    struct string * returned_name = NULL;
	    struct string * official_name = 0;
	    const struct string *require_name = NULL;
	    int found = 0;

	    if ((ra->service->flags & SE_require_tls_name)) {
		start = -1;
		official_name = new_string2(system_charset,
					    s2us(ra->service->
						 official_name));
	    }

	    for (i = start; i < ra->service->req_tls_peername_count; i++) {
		int r;

		if (i < 0)
		    require_name = official_name;
		else
		    require_name =  ra->service->req_tls_peername_list[i];

		r = StreamVerifyName(ra->stream,SS_peer_cn,require_name,
				     & returned_name);

		if (r) {
		    found = 1;
		    break;
		}		    
	    }
	 
	    if (!found) {
		ret = 0;

		if (start < ra->service->req_tls_peername_count -1 ||
		    !require_name) {
		    /* Several possible names */

		    if (returned_name) {
			lib_error(CATGETS(elm_msg_cat, MeSet, 
					  MeRACertCNWrongOther,
					  "%S %s certificate CN is %S. Other certificate is required."),
				  Server,ra->host,returned_name);
		    } else {
			int bits;
		
			StreamInfo(ra->stream,SS_ssf,&bits,NULL,NULL);

			if (bits > 0)
			    lib_error(CATGETS(elm_msg_cat, MeSet, 
					      MeRACertCNWrongCert,
					      "%S %s certificate CN is wrong."),
				      Server,ra->host);
			else
			    lib_error(CATGETS(elm_msg_cat, MeSet, 
					      MeRANotEncryptedCN,
					      "%S %s connection not encrypted, certificate required."),
				      Server,ra->host);
		    }

		} else {
		    if (returned_name) {
			lib_error(CATGETS(elm_msg_cat, MeSet, 
					  MeRACertCNWrongRequired,
					  "%S %s certificate CN is %S, but %S required."),
				  Server,ra->host,returned_name,
				  require_name);
		    } else {
			int bits;
			
			StreamInfo(ra->stream,SS_ssf,&bits,NULL,NULL);

			if (bits > 0)
			    lib_error(CATGETS(elm_msg_cat, MeSet, 
					      MeRACertCNRequired,
					      "%S %s certificate CN %S required."),
				      Server,ra->host,require_name);
			else
			    lib_error(CATGETS(elm_msg_cat, MeSet, 
					      MeRANotEncryptedRequired,
					      "%S %s connection not encrypted, but certificate CN %S required."),
				      Server,ra->host,require_name);
		    }

		}
	    }
	  
	    if (returned_name)			       
		    free_string(&returned_name);
	    if (official_name)
		free_string(&official_name);
	}	
    }

 failure:
    DPRINT(Debug,8,(&Debug,"remote_account_verify_peer=%d%s\n",
		    ret,
		    ret > 0 ? " (succeed)" : ""));

    return ret;
}

enum itls_status remote_account_init_tls(ra) 
     struct remote_account *ra;
{
    enum itls_status ret = itls_none;

    if (REMOTE_ACCOUNT_magic != ra->magic) {
	panic("CONNECTION PANIC",__FILE__,__LINE__,
	      "remote_account_init_tls",
	      "Bad magic (remote_account)",0);
    }

    if (ra->service) {

	const char * init_tls = NULL;

	if (SERVICE_ENTRY_magic != ra->service->magic)
	    panic("CONNECTION PANIC",__FILE__,__LINE__,
		  "remote_account_init_tls",
		  "Bad magic (service_entry)",0);

	init_tls =  have_init_tls(ra->service);

	if (init_tls) {
#ifdef USE_DLOPEN
	    ret = initial_TLS_ra(ra,init_tls);
#else
	    DPRINT(Debug,9,(&Debug,
			    "remote_account_init_tls:  no shared routines available\n"));
	    ret = itls_unsupported;
#endif

	    if (itls_unsupported == ret) {
		lib_error(CATGETS(elm_msg_cat, MeSet,MeConTLSUnsupported,
				  "%s connection to %s does not support SSL/TLS."),
			  init_tls,ra->host);

	    }
	}
    }

    
    return ret;
}

#endif
    
/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
